﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;


namespace GraphBundling2D
{
    /// <summary>
    /// 
    /// Code :
    ///     Antoine LHUILLIER 
    ///     Christophe HURTER
    ///     
    /// Reference:
    /// C. Hurter, O. Ersoy, and A. Telea. 2012.
    /// Graph Bundling by Kernel Density Estimation. 
    /// Comp. Graph. Forum 31, 3pt1 (June 2012), 865-874.
    /// DOI=10.1111/j.1467-8659.2012.03079.x http://dx.doi.org/10.1111/j.1467-8659.2012.03079.x 
    /// 
    /// Abstract:
    /// We present a fast and simple method to compute bundled layouts of general graphs.
    /// For this, we first transform a given graph drawing into a density map using kernel density estimation.
    /// Next, we apply an image sharpening technique which progressively merges local height maxima by moving
    /// the convolved graph edges into the height gradient flow. Our technique can be easily and efficiently 
    /// implemented using standard graphics acceleration techniques and produces graph bundlings of similar 
    /// appearance and quality to state-of-the-art methods at a fraction of the cost. Additionally,
    /// we show how to create bundled layouts constrained by obstacles and use shading to convey information
    /// on the bundling quality. We demonstrate our method on several large graphs. © 2012 Wiley Periodicals, Inc.
    /// 
    /// </summary>
    public partial class Bundling2DForm : Form
    {
        #region Attribute

        // Set of current, original trajectories(organized records)
        private Database dataSet;
        private float[] SplattedAccMap;

        // Color define 
        private Color BackgroundColor = Color.White;

        // Pan and zoom parameters
        private double PanX = 1;
        private double PanY = 1;
        private double Zoom = 0.5;
     
        // View parameters
        private bool ShowPoints = true;

        // Sample Parameters
        private double splitDistance = 0.005;
        private double removeDistance = 0.0025;

        // Splatting & Accumulation Parameters
        private int AccResolution = 300;
        private static int KernelSize = 11;
        private static float[] Kernel = GetKernel(KernelSize);

        // Gradient Application Parameters
        private int GradientW = 15;
        private double attractionFactor = 1.0;

        //Smooth parameters
        int smoothValue = 50;
        #endregion

        #region Initialization

        public Bundling2DForm()
        {
            InitializeComponent();

            // Initialize the dataset
            dataSet = new Database();
            PictureBoxBundling.Paint += new PaintEventHandler(GLControl_Paint);
        }

        #endregion

        #region BundelButton_Click

        private void bundle_Click(object sender, EventArgs e)
        {
            // Resampling
            Resample(splitDistance, removeDistance);

            // Splatting
            ComputeSplatting();

            // Apply Gradient
            ApplyGradient(GradientW, attractionFactor, SplattedAccMap);

            // Smooth trajectories
            for (int i = 0; i < smoothValue; i++)
            {
                SmoothTrajectories();
            }

            // Update the view
            PictureBoxBundling.Invalidate();
        }

        #endregion

        #region Full Bundle Button
        // Do 20 bundle in a row
        private void bundle20Button_Click(object sender, EventArgs e)
        {
            for (int i = 0; i < 20; i++)
            {
                bundle_Click(sender, e);
            }
        }
        #endregion

        #region DrawData

        private void DrawData(Graphics g)
        {
            // Draw records trajectories
            Vertex[][] tempVertex = dataSet.getCurrentTraj();
            foreach (var lines in tempVertex)
            {
                foreach (var pt in lines)
                {
                    for (int i = 0; i < lines.Length - 1; i++)
                    {
                        PointF src1 = GetScreenCoordinates((float)(2.0 * lines[i].getX() - 1.0), (float)(2.0 * lines[i].getY() - 1.0));
                        PointF src2 = GetScreenCoordinates((float)(2.0 * lines[i + 1].getX() - 1.0), (float)(2.0 * lines[i + 1].getY() - 1.0));
                        g.DrawLine(Pens.Black, src1, src2);
                    }
                }
            }

            // Draw Points
            if (ShowPoints)
            {
                float pointSize = 1.0f + (float)Zoom;
                foreach (var lines in tempVertex)
                {
                    foreach (var pt in lines)
                    {
                        PointF src = GetScreenCoordinates((float)(2.0 * pt.getX() - 1.0), (float)(2.0 * pt.getY() - 1.0));
                        g.FillEllipse(Brushes.Red, src.X, src.Y, pointSize, pointSize);

                    }

                }
            }
        }

        #endregion

        #region Sampling

        /// <summary>
        /// Sampling extrapolate a new point from two point if thoose are too far from each other.
        /// But also can suppress a point if he is to close to another.
        /// Sampling is done only one vertices from the same trajectory.
        /// </summary>
        /// <param name="splitDist"></param>
        /// <param name="removeDist"></param>
        private void Resample(double splitDist, double removeDist)
        {
            Vertex[][] tempVertex = dataSet.getCurrentTraj();
            for (int i = 0; i < tempVertex.Length; i++)
            {
                List<Vertex> tmpVertexList = new List<Vertex>();
                tmpVertexList.Add(tempVertex[i][0]);
                // Sampling on each lines (= trajectory)
                for (int j = 0; j < tempVertex[i].Length - 1; j++)
                {
                    Vertex currentVert = tempVertex[i][j];
                    Vertex nextVert = tempVertex[i][j + 1];
                    double dist = Math.Sqrt((currentVert.getX() - nextVert.getX()) * (currentVert.getX() - nextVert.getX())
                        + (currentVert.getY() - nextVert.getY()) * (currentVert.getY() - nextVert.getY()));
                    // Test if the next point is too far or too close
                    if (dist > splitDist)
                    {
                        tmpVertexList.Add(new Vertex((currentVert.getX() + nextVert.getX()) / 2.0, (currentVert.getY() + nextVert.getY()) / 2.0, (currentVert.getZ() + nextVert.getZ()) / 2.0, currentVert.getColor()));
                    }
                    if (!(dist < removeDist) || j == tempVertex[i].Length - 2)
                    {
                        tmpVertexList.Add(nextVert);
                    }
                }
                tempVertex[i] = tmpVertexList.ToArray();
            }
            dataSet.CurrentTraj = tempVertex;
        }

        #endregion

        #region KernelWrapper

        private static float[] GetKernel(int KernSize)
        {
            // Creates the kernel Structure
            float[] Kernel = new float[KernSize * KernSize];

            // Fill the structure for a circular gradiant Kernel
            for (int i = 0; i < KernSize * KernSize; i += KernSize)
            {
                for (int j = 0; j < KernSize; j++)
                {
                    float centerDist = (float)Math.Sqrt(((float)j - ((float)KernSize - 1.0f) / 2.0f) * ((float)j - ((float)KernSize - 1.0f) / 2.0f)
                        + ((float)i / ((float)KernSize) - ((float)KernSize - 1.0f) / 2.0f) * ((float)i / ((float)KernSize) - ((float)KernSize - 1.0f) / 2.0f));
                    Kernel[i + j] = Math.Max(0.0f, 1.0f - Math.Abs(2.0f / (float)(KernSize - 1) * centerDist));
                }
            }
            return Kernel;
        }

        #endregion

        #region Splatting method 2

        private void ComputeSplatting()
        {
            float[] AccMap = new float[AccResolution * AccResolution];
            foreach (var line in dataSet.CurrentTraj)
            {
                foreach (var point in line)
                {
                    for (int i = 0; i < Kernel.Length; i += KernelSize)
                    {
                        for (int j = 0; j < KernelSize; j++)
                        {
                            if (Convert.ToInt32(point.getX() * (double)AccResolution) * AccResolution + (i / KernelSize - KernelSize / 2) * AccResolution > 0
                                && Convert.ToInt32(point.getX() * (double)AccResolution) * AccResolution + (i / KernelSize - KernelSize / 2) * AccResolution < AccMap.Length
                                && Convert.ToInt32(point.getY() * (double)AccResolution) + (j - KernelSize / 2) > 0
                                && Convert.ToInt32(point.getY() * (double)AccResolution) + (j - KernelSize / 2) < AccResolution)
                            {
                                AccMap[Convert.ToInt32(point.getX() * (double)AccResolution) * AccResolution + Convert.ToInt32(point.getY() * (double)AccResolution) + (i / KernelSize - KernelSize / 2) * AccResolution + (j - KernelSize / 2)] += Kernel[i + j];
                            }
                        }
                    }
                }
            }
            SplattedAccMap = AccMap;
        }

        #endregion

      
        #region ApplyGradient for all points

        private void ApplyGradient(int gradientW, double attracFactor, float[] AccMap)
        {
            // Apply the Gradient for each points
            foreach (var lines in dataSet.CurrentTraj)
            {
                for (int j = 1; j < lines.Length - 1; j++)
                {
                    // Get the gradient value at the current point
                    int pointIndex = Convert.ToInt32(lines[j].getX() * (double)AccResolution) * AccResolution + Convert.ToInt32(lines[j].getY() * (double)AccResolution);
                    Vector2 localGradient = GetLocalGradient(gradientW, AccMap, lines[j], pointIndex);

                    // Normalizing the localGradient.
                    if (localGradient.X != 0 || localGradient.Y != 0)
                    {
                        localGradient.Normalize();
                    }

                    // Move each point by the gradient vector associated within
                    lines[j].setX(lines[j].getX() + attracFactor * localGradient.X / (double)AccResolution);
                    lines[j].setY(lines[j].getY() + attracFactor * localGradient.Y / (double)AccResolution);
                }
            }
        }

        #endregion

        #region GetLocalGradient

        private Vector2 GetLocalGradient(int gradientW, float[] AccMap, Vertex pt, int pointIndex)
        {
            Vector2 localGradient = new Vector2();
            for (int i = 0; i < gradientW; i++)
            {
                for (int j = 0; j < gradientW; j++)
                {
                    int offsetIndex = (i - GradientW / 2) * AccResolution + j - GradientW / 2;
                    if ((pointIndex + offsetIndex) / AccResolution > 0
                        && (pointIndex + offsetIndex) / AccResolution < AccResolution
                        && (pointIndex + offsetIndex) % AccResolution > 0
                        && (pointIndex + offsetIndex) % AccResolution < AccResolution)
                    {
                        float localDensity = AccMap[pointIndex + offsetIndex];
                        float dX = i - (GradientW) / 2;
                        float dY = j - (GradientW) / 2;
                        localGradient.X += localDensity * dX;
                        localGradient.Y += localDensity * dY;
                    }
                }
            }
            return localGradient;
        }

        #endregion

        #region Smooth trajectories

        private void SmoothTrajectories()
        {
            foreach (var lines in dataSet.CurrentTraj)
            {
                for (int i = 1; i < lines.Length - 1; i++)
                {
                    lines[i].setX((lines[i - 1].getX() + lines[i].getX() + lines[i + 1].getX()) / 3.0);
                    lines[i].setY((lines[i - 1].getY() + lines[i].getY() + lines[i + 1].getY()) / 3.0);
                }
            }
        }

        #endregion

        #region GLControl_Handlers
      
        #region Reference change methods

        #region OpenTKMousePosition

        PointF GetOpenTKMousePosition(Point mouseScreenPos)
        {
            PointF pt = new PointF();
            pt.X = ((float)mouseScreenPos.X / (float)(PictureBoxBundling.Width - 1) * 2 - 1);
            pt.Y = ((float)((PictureBoxBundling.Height - 1) - mouseScreenPos.Y) / (float)(PictureBoxBundling.Height - 1) * 2 - 1);
            return pt;
        }

        #endregion

        #region GetScreenCoordinates

        PointF GetScreenCoordinates(float Xi, float Yi)
        {
            return new PointF((float)Zoom * (Xi + (float)PanX) * PictureBoxBundling.Width, (float)Zoom * (Yi + (float)PanY) * PictureBoxBundling.Height);
        }

        #endregion

        #region GetImageCoordinates

        PointF GetImageCoordinates(float Xs, float Ys)
        {
            return new PointF(Xs / (float)Zoom - (float)PanX, Ys / (float)Zoom - (float)PanY);
        }

        #endregion

        #endregion

        #region Paint
        private void GLControl_Paint(object sender, PaintEventArgs e)
        {
            DrawData(e.Graphics);
        }
        #endregion

        #region Show Point CheckBox
        private void ShowPointscheckBox_CheckedChanged(object sender, EventArgs e)
        {
            if (ShowPoints)
            {
                ShowPoints = false;
            }
            else
            {
                ShowPoints = true;
            }
            PictureBoxBundling.Invalidate();
        }
        #endregion

     
        #endregion
    }
}

